%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Db.h>
#include <DbDefs.h>
#include <DbObject.h>
#include <DbParse.h>

typedef struct
{
  char* buffer;
  int length;
  int current_position;
  int line_number;
  char* old_line_start;
  char* line_start;
} LexContext;

int DbLexLineNumber ();
char* DbLexLineStart ();
int DbLexPos ();
LexContext* DbLexStartup (LexContext* context, char* text);

#ifdef YYLMAX
#undef YYLMAX
#endif

#define YYLMAX 2000

int DbLexInput ();
void DbLexUnput (int c);

#ifndef YY_NO_INPUT

#undef input
#define input DbLexInput
#undef unput
#define unput DbLexUnput

#else

#define internal_input DbLexInput

#undef YY_INPUT
#define YY_INPUT(buffer,result,max_size) \
{ \
  int c = '*', n; \
  for ( n = 0; n < max_size && \
               ((c = internal_input()) != 0) && \
               (c != '\n'); ++n ) \
    { \
      buffer[n] = (char) c; \
    } \
  if ( c == '\n' ) buffer[n++] = (char) c; \
  result = n; \
}

#endif

static LexContext* Context = NULL;

static char* StackPut (char* text);

int DbYYlex ();

%}

%start DB
%start DBCOMMENT

space     [ ;\t\r\f]
alphanum  [a-zA-Z0-9][a-zA-Z0-9_\-]*
term      [^a-zA-Z0-9]

%%

<DB>"#" {
  BEGIN DBCOMMENT;
}

<DBCOMMENT>\n {
  Context->old_line_start = Context->line_start;
  Context->line_start = &Context->buffer[Context->current_position];
  Context->line_number++;

  BEGIN DB;
}

<DBCOMMENT>. {
}

<DB>\n {
  Context->old_line_start = Context->line_start;
  Context->line_start = &Context->buffer[Context->current_position];
  Context->line_number++;
}

<DB>{space}+ {
}

<DB>"{" {
  return (OPEN_BRACE);
}

<DB>"}" {
  return (CLOSE_BRACE);
}

<DB>"[" {
  return (OPEN_SQUARE);
}

<DB>"]" {
  return (CLOSE_SQUARE);
}

<DB>"<" {
  return (LOWER);
}

<DB>">" {
  return (GREATER);
}

<DB>"(" {
  return (OPEN_PAR);
}

<DB>")" {
  return (CLOSE_PAR);
}

<DB>"," {
  return (COMMA);
}

<DB>"=" {
  return (EQUAL);
}

<DB>"/" {
  return (SLASH);
}

<DB>"short" {
  return (SHORT);
}

<DB>"int" {
  return (INT);
}

<DB>"hexa" {
  return (HEXA);
}

<DB>"float" {
  return (FLOAT);
}

<DB>"double" {
  return (DOUBLE);
}

<DB>"string" {
  return (STRING);
}

<DB>"reference" {
  return (REFERENCE);
}

<DB>"char" {
return (CHARACTER);
}

<DB>"version" {
return (VERSION);
}

<DB>"no"{space}*"version" {
return (NOVERSION);
}

<DB>"@" {
return (AT);
}

<DB>[0][Xx][0-9a-fA-F][0-9a-fA-F]*/{term} {
  int n;

  sscanf ((char*) &yytext[2], "%x", &n);

  yylval.number = n;
  return (NUMBER);
}

<DB>([+-])?[0-9][0-9]*\.[0-9]*([Ee]([+-])?[0-9][0-9]*)?/{term} {
  double d;

  d = strtod ((const char *) yytext, NULL);

  yylval.real = d;
  return (REAL);
}

<DB>([+-])?\.[0-9][0-9]*([Ee]([+-])?[0-9][0-9]*)?/{term} {
  double d;

  d = strtod ((const char *) yytext, NULL);

  yylval.real = d;
  return (REAL);
}

<DB>([+-])?[0-9][0-9]*[Ee]([+-])?[0-9][0-9]*/{term} {
  double d;

  d = strtod ((const char *) yytext, NULL);

  yylval.real = d;
  return (REAL);
}

<DB>([+-])?[0-9][0-9]*/{term} {
  int n;

  sscanf ((char*) yytext, "%d", &n);

  yylval.number = n;
  return (NUMBER);
}

<DB>\"[^"]*\"/{term} {
/*"*/
  char* string;

  yytext[yyleng-1] = 0;
  string = (char*) &yytext[1];

  yylval.string = StackPut (string);

  return (TEXT);
}

<DB>\'[^']*\'/{term} { /*'*/
  char c;

  sscanf ((char*) &yytext[1], "%c", &c);

  yylval.number = c;

  return (NUMBER);
}

<DB>{alphanum} {
  char* string;

  string = (char*) &yytext[0];

  yylval.string = StackPut (string);

  return (NAME);
}

<DB>. {
  DbSetErrorInfo (DbErrorBadSyntax, 
                  "Unknown syntax %s\n", (char*) &yytext[0]);
  DbPrintError ();
}

%%

/*---------------------------------------------------------------*/
int DbLexLineNumber ()
/*---------------------------------------------------------------*/
{
  return (Context->line_number);
}


/*---------------------------------------------------------------*/
char* DbLexLineStart ()
/*---------------------------------------------------------------*/
{
  return (Context->line_start);
}

/*---------------------------------------------------------------*/
int DbLexPos ()
/*---------------------------------------------------------------*/
{
  int pos;

  pos = Context->line_start - &Context->buffer[Context->current_position];

  return (pos);
}

/*---------------------------------------------------------------*/
int DbLexInput ()
/*---------------------------------------------------------------*/
{
  char c;

  if (!Context->buffer) return (0);

  if (Context->current_position < Context->length)
    {
      c = Context->buffer[Context->current_position];
      Context->current_position++;
    }
  else c = 0;

  return (c);
}

/*---------------------------------------------------------------*/
void DbLexUnput (int c)
/*---------------------------------------------------------------*/
{
  if (!Context->buffer) return;

  if (Context->current_position > 0)
  {
    Context->current_position--;
    Context->buffer[Context->current_position] = c;
  }
}

/*---------------------------------------------------------------*/
void DbLexRestoreContext (LexContext* context)
/*---------------------------------------------------------------*/
{
  if (Context->buffer != NULL)
    {
      free (Context->buffer);
      Context->buffer = NULL;
    }

  Context = context;
}

/*---------------------------------------------------------------*/
LexContext* DbLexStartup (LexContext* context, char* text)
/*---------------------------------------------------------------*/
{
  LexContext* previous = Context;

  Context = context;

  yyin = stdin;
  yyout = stdout;

  Context->buffer           = NULL;
  Context->length           = 0;
  Context->current_position = 0;
  Context->line_number      = 0;
  Context->old_line_start   = NULL;
  Context->line_start       = NULL;

  Context->buffer = (char*) malloc (strlen (text) + 2);
  strcpy (Context->buffer, text);
  strcat (Context->buffer, " ");

  Context->length = strlen (Context->buffer);

  Context->current_position = 0;
  Context->line_number = 1;
  Context->old_line_start = Context->buffer;
  Context->line_start = Context->buffer;

  BEGIN DB;

  return (previous);
}

/*---------------------------------------------------------------*/
static char* StackPut (char* text)
/*---------------------------------------------------------------*/
{
#define MAXDEPTH 150

  static struct
  {
    char text[YYLMAX];
  } stack[MAXDEPTH];

  static int index = 0;

  char* str;

  str = (char*) stack[index].text;
  strcpy (str, text);

  index++;

  if (index >= MAXDEPTH) 
  index = 0;

  return (str);
}
