/*---------------------------------------------------------------------------*/
/*        Copyright (c) 1996 LAL Orsay, UPS-IN2P3-CNRS (France).             */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/* 1. Redistributions of source code must retain the above copyright         */
/*    notice, this list of conditions and the following disclaimer.          */
/* 2. Redistributions in binary form must reproduce the above copyright      */
/*    notice, this list of conditions and the following disclaimer in the    */
/*    documentation and/or other materials provided with the distribution.   */
/* 3. All advertising materials mentioning features or use of this software  */
/*    must display the following acknowledgement:                            */
/*      This product includes software developed by the Computer Application */
/*      Development Group at LAL Orsay (Laboratoire de l'Accelerateur        */
/*      Linaire - UPS-IN2P3-CNRS).                                           */
/* 4. Neither the name of the Institute nor of the Laboratory may be used    */
/*    to endorse or promote products derived from this software without      */
/*    specific prior written permission.                                     */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE LAL AND CONTRIBUTORS ``AS IS'' AND       */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE     */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR        */
/* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE LAL OR CONTRIBUTORS BE      */
/* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR       */
/* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF      */
/* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS  */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN   */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)   */
/* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF    */
/* THE POSSIBILITY OF SUCH DAMAGE.                                           */
/*---------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <CSet.h>
#include <DbDefs.h>
#include <DbObjectPath.h>
#include <DbObjectP.h>

/*---------------------------------------------------------------------------*/
static void DbObjectDelete (DbObject object);
static void DbObjectErase (DbObject object);
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
static CHasher Objects = NULL;
static GCClass DbObjectGCClass = NULL;
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
DbObject DbObjectFind (char* name)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbObjectFindInPath (name, NULL);

  return (object);
}

/*---------------------------------------------------------------------------*/
DbObject DbObjectFindInPath (char* name, char* path_name)
/*---------------------------------------------------------------------------*/
{
  DbObject object;
  DbObjectPath path;

  path = DbObjectPathFind (path_name);
  if (path == NULL) return (NULL);

  object = DbObjectPathFindObject (path, name);

  return (object);
}

/*---------------------------------------------------------------------------*/
void DbObjectCleanup ()
/*---------------------------------------------------------------------------*/
{
  DbObjectPathCleanup ();
  GCReferenceClear (DbObjectGCClass);

  GCCleanup ();
}

/*---------------------------------------------------------------------------*/
DbObject DbObjectNewInPath (DbType type, char* name, char* path_name)
/*---------------------------------------------------------------------------*/
/*                                                                           */
/*   Cree une copie de l'objet.                                              */
/*  Il est de la responsabilite de l'appelant de verifier si l'objet         */
/*  existe deja (est deja installe) ou si on veut une copie temporaire.      */
/*                                                                           */
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  DbObjectPathNew (path_name);

  object = (DbObject) malloc (sizeof (DbObjectRec));
  object->type = NULL;

  object->name = (char*) malloc (strlen (name) + 1);
  strcpy (object->name, name);

  if (path_name == NULL) path_name = "";
  object->path_name = (char*) malloc (strlen (path_name) + 1);
  strcpy (object->path_name, path_name);

  object->path  = NULL;
  object->value = NULL;
  object->versioning_option = DbDefaultVersioning;
  object->owner = NULL;

  if (DbObjectGCClass == NULL)
    {
      GCReferenceAssign (DbObjectGCClass,
                         GCClassNew ("DbObject", 
                                     (GCDeleteMethod) DbObjectDelete));
    }

  GCObjectNew (object, DbObjectGCClass);

  DbObjectInitialize (object, type);

  return (object);
}

/*---------------------------------------------------------------------------*/
int DbObjectInstall (DbObject object)
/*---------------------------------------------------------------------------*/
/*                                                                     */
/*   Returns whether a true installation occured (1)                   */
/*  or (0) if the object was already installed.                        */
/*                                                                     */
/*---------------------------------------------------------------------*/
{
  if (object == NULL) return (0);

  return (DbObjectInstallInPath (object, object->path_name));
}

/*---------------------------------------------------------------------------*/
int DbObjectInstallInPath (DbObject object, char* path_name)
/*---------------------------------------------------------------------------*/
/*                                                                     */
/*   Returns whether a true installation occured (1)                   */
/*  or (0) if the object was already installed.                        */
/*                                                                     */
/*---------------------------------------------------------------------*/
{
  DbObjectPath path;

  if (object == NULL) return (0);

  if (path_name == NULL) path_name = "";

  if (DbObjectFindInPath (DbObjectGetName (object), path_name) != NULL) 
    return (0);

  path = DbObjectPathFind (path_name);
  if (path == NULL) return (0);

  DbObjectPathInstallObject (path, object);

  if ((path_name != object->path_name) &&
      (strcmp (object->path_name, path_name) != 0))
    {
      free (object->path_name);
      object->path_name = NULL;

      object->path_name = (char*) malloc (strlen (path_name) + 1);
      strcpy (object->path_name, path_name);
    }

  GCReferenceAssignIn (object->path, object, path);

  return (1);
}

/*---------------------------------------------------------------------------*/
void DbObjectRemove (DbObject object)
/*---------------------------------------------------------------------------*/
{
  DbObjectPath path;

  if (object == NULL) return;

  path = object->path;
  if (path == NULL) return;

  DbObjectPathRemoveObject (path, object);
  DbObjectDestroy (object);

  GCReferenceClear (object->path);

  /*GCCleanup ();*/
}

/*---------------------------------------------------------------------------*/
static void DbObjectErase (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return;

  if (object->value != NULL)
    {
      DbTypeErase (object->type, object->value);

      free (object->value);
      object->value = NULL;
    }
}

/*---------------------------------------------------------------------------*/
void DbObjectClear (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return;

  if (object->value != NULL)
    {
      DbTypeErase (object->type, object->value);
      DbTypeInit (object->type,  object->value);
    }
}

/*---------------------------------------------------------------------------*/
void DbObjectInitialize (DbObject object, DbType type)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return;

  DbObjectErase (object);

  GCReferenceAssignIn (object->type, object, type);

  object->value = DbTypeInstanciate (object->type);
}

/*---------------------------------------------------------------------------*/
void DbObjectDestroy (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return;

  DbObjectErase (object);

  GCReferenceClear (object->type);
}

/*---------------------------------------------------------------------------*/
void DbObjectsDestroy (CLList objects)
/*---------------------------------------------------------------------------*/
{
  CIterator it;
  
  if (objects == NULL) return;

  it = CLListGiveIterator (objects);

  for (; !CIteratorFinished (it); CIteratorForth (it))
  {
    DbObject object = NULL;

    object = (DbObject) CIteratorItem (it);

    if (object != NULL)
    {
      DbObjectDestroy (object);
    }
  }

  CIteratorDelete  (it);
}



/*---------------------------------------------------------------------------*/
static void DbObjectDelete (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return;

  DbObjectDestroy (object);

  if (object->name != NULL)
    {
      free (object->name);
      object->name = NULL;
    }

  if (object->path_name != NULL)
    {
      free (object->path_name);
      object->path_name = NULL;
    }

  if (object->owner != NULL)
    {
      free (object->owner);
      object->owner = NULL;
    }

  free (object);
}

/*---------------------------------------------------------------------------*/
void DbObjectSetOwner (DbObject object, char* owner)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return;

  if (object->owner != NULL)
    {
      free (object->owner);
      object->owner = NULL;
    }

  if ((owner != NULL) &&
      (strlen (owner) > 0))
    {
      object->owner = (char*) malloc (strlen (owner) + 1);
      strcpy (object->owner, owner);
    }
}

/*---------------------------------------------------------------------------*/
DbType DbObjectGetType (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return (NULL);

  return (object->type);
}

/*---------------------------------------------------------------------------*/
char* DbObjectGetName (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return (NULL);

  return (object->name);
}

/*---------------------------------------------------------------------------*/
char* DbObjectGetOwner (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return (NULL);

  return (object->owner);
}

/*---------------------------------------------------------------------------*/
DbVersioningOption DbObjectGetVersioningOption (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return (DbDefaultVersioning);

  return (object->versioning_option);
}

/*---------------------------------------------------------------------------*/
void DbObjectSetVersioningOption (DbObject object, DbVersioningOption option)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return;

  object->versioning_option = option;
}

/*---------------------------------------------------------------------------*/
char* DbObjectGetPathName (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return (NULL);

  return (object->path_name);
}

/*---------------------------------------------------------------------------*/
DbTypeIterator DbObjectGiveInputIterator (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return (NULL);

  return (DbTypeGiveInputIterator (object->type, object->value));
}

/*---------------------------------------------------------------------------*/
DbTypeIterator DbObjectGiveOutputIterator (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return (NULL);

  return (DbTypeGiveOutputIterator (object->type, object->value));
}

/*- procedure ---------------------------------------------------------*/
void DbObjectVisit (DbObject object, DbVisitor visitor)
/*---------------------------------------------------------------------*/
{
  DbType type;
  DbTypeIterator it;

  if (object == NULL) return;
  if (visitor == NULL) return;

  type = DbObjectGetType (object);
  it = DbObjectGiveInputIterator (object);

  for (;;)
    {
      if (!DbTypeIteratorFinished (it))
        {
          DbTypeIterator sub;

          sub = DbTypeIteratorDown (it);
          if (sub != NULL)
            {
              if (DbVisitorDown (visitor, it))
                {
                  it = sub;
                }
              else
                {
                  DbTypeIteratorUp (sub);
                }
            }
          else
            {
              if (DbVisitorIn (visitor, it))
                {
                  DbTypeIteratorForth (it);
                }
              else
                {
                  DbTypeIterator parent;

                  parent = DbTypeIteratorUp (it);
                  if (parent == NULL)
                    {
                      break;
                    }
                  else
                    {
                      DbVisitorUp (visitor, it);
                      it = parent;
                      DbTypeIteratorForth (it);
                    }
                }
            }
        }
      else
        {
          DbTypeIterator parent;

          parent = DbTypeIteratorUp (it);
          if (parent == NULL)
            {
              break;
            }
          else
            {
              DbVisitorUp (visitor, it);
              it = parent;
              DbTypeIteratorForth (it);
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void* DbObjectGetAddress (DbObject object, DbType type)
/*---------------------------------------------------------------------------*/
{
  if (object->type != type) return (NULL);

  return ((void*) object->value);
}

/*---------------------------------------------------------------------------*/
char* DbCharObjectGetAddress (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (!DbTypeIsChar (object->type)) return (NULL);

  return ((char*) object->value);
}

/*---------------------------------------------------------------------------*/
short* DbShortObjectGetAddress (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (!DbTypeIsShort (object->type)) return (NULL);

  return ((short*) object->value);
}

/*---------------------------------------------------------------------------*/
int* DbIntObjectGetAddress (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (!DbTypeIsInt (object->type)) return (NULL);

  return ((int*) object->value);
}

/*---------------------------------------------------------------------------*/
unsigned int* DbHexaObjectGetAddress (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (!DbTypeIsHexa (object->type)) return (NULL);

  return ((unsigned int*) object->value);
}

/*---------------------------------------------------------------------------*/
float* DbFloatObjectGetAddress (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (!DbTypeIsFloat (object->type)) return (NULL);

  return ((float*) object->value);
}

/*---------------------------------------------------------------------------*/
double* DbDoubleObjectGetAddress (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (!DbTypeIsDouble (object->type)) return (NULL);

  return ((double*) object->value);
}

/*---------------------------------------------------------------------------*/
DbString* DbStringObjectGetAddress (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (!DbTypeIsString (object->type)) return (NULL);

  return ((DbString*) object->value);
}

/*---------------------------------------------------------------------------*/
DbVector* DbVectorObjectGetAddress (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (!DbTypeIsVector (object->type)) return (NULL);

  return ((DbVector*) object->value);
}

/*---------------------------------------------------------------------------*/
DbMatrix* DbMatrixObjectGetAddress (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (!DbTypeIsMatrix (object->type)) return (NULL);

  return ((DbMatrix*) object->value);
}

/*---------------------------------------------------------------------------*/
void* DbStructureObjectGetAddress (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (!DbTypeIsStructure (object->type)) return (NULL);

  return ((DbReference*) object->value);
}

/*---------------------------------------------------------------------------*/
DbReference* DbReferenceObjectGetAddress (DbObject object)
/*---------------------------------------------------------------------------*/
{
  if (!DbTypeIsReference (object->type)) return (NULL);

  return ((DbReference*) object->value);
}
