%{

/*-- Include files ----------------------------------------------*/
#include <stdarg.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define alloca malloc
/*---------------------------------------------------------------*/

/*---------------------------------------------------------------*/
#include <Db.h>
/*---------------------------------------------------------------*/


/*-- Prototypes -------------------------------------------------*/
#define yyparse DbTypeSemanticAnalyzer
#define yyerror DbTypeLexicalError
#define yyerrlab DbTypeErrLbl
#define yyerrflag DbTypeErrFlg

int DbTypeSemanticAnalyzer ();
int DbTypeYYlex ();

typedef enum
{
  TypeDefinition,
  TypeRedefinition
} ParserMode;

typedef struct
{
  ParserMode       mode;
  DbType           type;
  DbType           field_type;
  DbTypeIterator   type_iterator;
  CLList           lists;
} ParserContext;

typedef struct
{
  char* buffer;
  int current_position;
  int line_number;
  char* old_line_start;
  char* line_start;
} LexContext;

static ParserContext* Context = NULL;
static LexContext*    Lex_context = NULL;
 
int DbTypeLexLineNumber ();
char* DbTypeLexLineStart ();
int DbTypeLexPos ();
void DbTypeLexRestoreContext (LexContext* context);
LexContext* DbTypeLexStartup (LexContext* context, char* text);

static int DbTypeLexicalError(char* s);
static CLList AddFields (CLList list, CLList fields);
/*---------------------------------------------------------------*/
%}

%union
{
  DbField     field;
  CLList      list;
  DbObject    object;
  DbType      type;
  double      real;
  int         number;
  char*       string;
}

%start db

%token STRUCTURE
%token <string> NAME
%token OPEN_BRACE
%token CLOSE_BRACE
%token COMMA

%token SHORT
%token INT
%token HEXA
%token FLOAT
%token DOUBLE

%token CHARACTER
%token STRING

%token REFERENCE

%token OPEN_SQUARE
%token CLOSE_SQUARE
%token LOWER
%token GREATER

%token <number> NUMBER

%type <type> defstructure
%type <type> basic_field_type
%type <type> field_set
%type <list> opt_fields
%type <list> fields
%type <list> field
%type <list> field_names

%%

db :
  /* empty */
  | defstructure
    {
      DbType type = $1;
    }
  ;

defstructure :
    STRUCTURE NAME 
    {
      char* name = $2;
      DbType type = NULL;

      if (Context->mode == TypeDefinition)
        {
          type = DbTStructureFind (name);
              
          if (type == NULL)
            {
              Context->type = DbTStructureNew (name, NULL);
              DbTStructureInstall (Context->type);
            }
          else
            {
              DbSetErrorInfo (DbErrorTypeAlreadyDefined, 
                              "Type %s already defined", 
                              name);

              Context->type = NULL;
            }
        }
      else
        {
          type = DbTStructureFind (name);
              
          if (type == NULL)
            {
              type = DbTStructureNew (name, NULL);
              DbTStructureInstall (type);
            }

          Context->type = type;
        }
    }
    OPEN_BRACE opt_fields CLOSE_BRACE
    {
      char* name = $2;
      CLList fields = $5;
      DbType type = Context->type;

      if (fields != NULL)
        {
          if (type != NULL)
            {
              DbType new_type = NULL;
                  
              GCReferenceAssign (new_type, DbTStructureNew (name, fields));
                  
              if (new_type != type)
                {
                    /*
                      This situation occurs when an already defined
                      structure (ie. same name) has been redefined
                      (ie. its field list has changed).
                      
                      We must convert all objects and types using 
                      directly or indirectly this structure.
                      
                      Eventually 'type' will be deleted and 'new_type'
                      installed in place of it in the data base.
                      */
                  
                    /*
                      to be done...
                      
                      There will be something like :
                      
                      DbTypeConvert (type, new_type);
                      
                      */
                  CLListErase (fields);
                }
                  
              GCReferenceClear (new_type);
            }
          else
            {
              CLListErase (fields);
            }

          CLListRemove (Context->lists, fields);
        }

      $$ = type;
    }
  ;

opt_fields :
  /* empty */
    {
      $$ = NULL;
    }
  | fields
    {
      $$ = $1;
    }
  ;

fields :
    field
    {
      CLList fields = $1;
      CLList result = NULL;

      result = AddFields (result, fields);

      $$ = result;
    }
  | fields field
    {
      CLList result = $1;
      CLList fields = $2;

      if ((fields != NULL) &&
          (result != NULL))
        {
          result = AddFields (result, fields);
        }
      else
        {
          if (fields != NULL)
            {
              CLListErase (fields);
              CLListRemove (Context->lists, fields);
              fields = NULL;
            }

          if (result != NULL)
            {
              CLListErase (result);
              CLListRemove (Context->lists, result);
              result = NULL;
            }
        }

      $$ = result;
    }
  ;

field :
    basic_field_type 
    {
      DbType type = $1;

      Context->field_type = type;
    }
    field_names
    {
      CLList list = $3;

      Context->field_type = NULL;
      $$ = list;
    }
  | field_set
    {
      DbType type = $1;

      Context->field_type = type;
    }
    field_names
    {
      CLList list = $3;

      Context->field_type = NULL;
      $$ = list;
    }
  ;

field_names :
    NAME
    {
      char* name = $1;
      CLList list = NULL;

      if ((name != NULL) &&
          (Context->field_type != NULL))
        {
          DbField field;

          list = CLListNew ();
          CLListAdd (Context->lists, list);

          field = DbFieldNew (Context->field_type, name);
          CLListAdd (list, field);
        }

      $$ = list;
    }
  | field_names COMMA NAME
    {
      CLList list = $1;
      char* name = $3;

      if ((name != NULL) &&
          (list != NULL))
        {
          DbField field;

          if (Context->field_type == NULL)
            {
              CLListErase (list);
              CLListRemove (Context->lists, list);
              list = NULL;
            }
          else
            {
              field = DbFieldNew (Context->field_type, name);
              CLListAdd (list, field);
            }
        }

      $$ = list;
    }
  ;

basic_field_type :
    CHARACTER
    {
      $$ = DbTCharNew ();
    }
  | SHORT
    {
      $$ = DbTShortNew ();
    }
  | INT
    {
      $$ = DbTIntNew ();
    }
  | HEXA
    {
      $$ = DbTHexaNew ();
    }
  | FLOAT
    {
      $$ = DbTFloatNew ();
    }
  | DOUBLE
    {
      $$ = DbTDoubleNew ();
    }
  | STRING
    {
      $$ = DbTStringNew ();
    }
  | REFERENCE
    {
      $$ = DbTReferenceNew (NULL);
    }
  | REFERENCE LOWER basic_field_type GREATER
    {
      DbType type = $3;

      $$ = DbTReferenceNew (type);
    }
  | NAME
    {
      char* name = $1;
      DbType type;

      type = DbTStructureFind (name);

      if (type == NULL)
        {
            /*
              Unknown user type
              */
          DbSetErrorInfo (DbErrorUnloadedStructure, "%s", name);
        }

      $$ = type;
    }
  | error
    {
      DbSetErrorInfo (DbErrorBadFieldSyntax, "DbParse>");
      DbTypeLexicalError ("Bad field basic type specification");

      $$ = NULL;
    }
  ;

field_set :
    basic_field_type OPEN_SQUARE CLOSE_SQUARE
    {
      DbType type = $1;

      $$ = DbTVectorNew (type);
    }
  | basic_field_type OPEN_SQUARE NUMBER CLOSE_SQUARE 
    OPEN_SQUARE NUMBER CLOSE_SQUARE
    {
      DbType type = $1;
      int rows = $3;
      int columns = $6;

      $$ = DbTMatrixNew (type, rows, columns);
    }
  | basic_field_type OPEN_SQUARE error CLOSE_SQUARE
    {
      DbType type = $1;

      DbSetErrorInfo (DbErrorBadDimension, "DbParse>");
      DbTypeLexicalError ("Bad field set type specification");

      $$ = NULL;
    }
  ;

%%

/*---------------------------------------------------------------*/
int yywrap ()
/*---------------------------------------------------------------*/
{
  return (1);
}

/*---------------------------------------------------------------*/
static int DbTypeLexicalError (char* s)
/*---------------------------------------------------------------*/
{
  int i;
  char* nl;
  char* start;
  int line_number;

  start = DbTypeLexLineStart ();
  line_number = DbTypeLexLineNumber ();

  fprintf (stderr, "DbType> %s at line %d :\n", s, line_number);

  nl = (char*) strchr (start, '\n');
  if (nl != NULL) *nl = 0;
  fprintf (stderr, "%s\n", start);
  if (nl != NULL) *nl = '\n';

  return (1);
}

/*---------------------------------------------------------------*/
static DbType DbTypeTextAnalyze (ParserMode mode, char* text)
/*---------------------------------------------------------------*/
{
  ParserContext context;
  ParserContext* previous_context = Context;
  LexContext* previous_lex_context;
  LexContext lex_context;
  DbType result = NULL;

  Context = &context;

  Context->mode          = mode;
  Context->type          = NULL;
  Context->field_type    = NULL;
  Context->type_iterator = NULL;
  Context->lists         = NULL;
  
  GCReferenceAssign (Context->lists, CLListNew ());
  
  previous_lex_context = DbTypeLexStartup (&lex_context, text);
  
  DbTypeSemanticAnalyzer ();
  DbTypeLexRestoreContext (previous_lex_context);
  
  CLListErase (Context->lists);
  GCReferenceClear (Context->lists);
  
  result = Context->type;
  
  Context = previous_context;
  
  return (result);
}

/*---------------------------------------------------------------*/
DbType DbParserCreateType (char* text)
/*---------------------------------------------------------------*/
{
  DbType type;

  type = DbTypeTextAnalyze (TypeDefinition, text);

  return (type);
}

/*---------------------------------------------------------------*/
DbType DbParserChangeType (char* text)
/*---------------------------------------------------------------*/
{
  DbType type;

  type = DbTypeTextAnalyze (TypeRedefinition, text);

  return (type);
}

/*---------------------------------------------------------------*/
static CLList AddFields (CLList list, CLList fields)
/*---------------------------------------------------------------*/
{
  CLList result = list;

  if (fields != NULL)
    {
      CIterator it;

      it = CLListGiveIterator (fields);
      for (; !CIteratorFinished (it); CIteratorForth (it))
        {
          DbField field;

          field = (DbField) CIteratorItem (it);

          if (result == NULL)
            {
              result = CLListNew ();
              CLListAdd (Context->lists, result);
            }
          CLListAdd (result, field);
        }
      
      CIteratorDelete (it);
      
      CLListErase (fields);
      CLListRemove (Context->lists, fields);
    }

  return (result);
}
